@@ -6,11 +6,14 @@ use std::{
 };
 
 use chrono::prelude::*;
+use ciborium::value::Value;
 use flate2::read::ZlibDecoder;
-use serde_cbor::Value;
 use serde_derive::Deserialize;
 use thiserror::Error;
 
+mod values;
+pub use values::*;
+
 type Result<T> = std::result::Result<T, Error>;
 
 #[derive(Deserialize)]
@@ -36,6 +39,9 @@ impl fmt::Display for Cwt {
 #[derive(Deserialize)]
 struct RawCert(BTreeMap<isize, Value>);
 
+#[derive(Deserialize)]
+struct RawHeader(BTreeMap<isize, Value>);
+
 /// Error type that represents every possible error condition encountered while loading a certificate
 #[derive(Debug, Error)]
 pub enum Error {
@@ -52,7 +58,7 @@ pub enum Error {
     InvalidFormatFor { key: String },
 
     #[error("failed to parse a payload as CBOR")]
-    MalformedCBOR(#[from] serde_cbor::Error),
+    MalformedCBOR(#[from] ciborium::de::Error<std::io::Error>),
 
     #[error("the root structure for the certificate is malformed")]
     MalformedCWT,
@@ -86,7 +92,7 @@ macro_rules! gen_extract {
     ($name:ident, $variant:path, $for_type:ty) => {
         fn $name(m: &mut BTreeMap<String, Value>, k: &str) -> Result<$for_type> {
             extract_key(m, k).and_then(|v| match v {
-                $variant(r) => Ok(r),
+                $variant(r) => Ok(r.into()),
                 _ => Err(Error::InvalidFormatFor { key: k.into() }),
             })
         }
@@ -211,7 +217,20 @@ impl TryFrom<BTreeMap<String, Value>> for GreenPass {
     }
 }
 
-/// Represents the whole certificate blob (excluding metadata and signature, which are unsupported at the moment)
+/// Represents the signature and signature metadata for a [HealthCert].
+#[derive(Debug, PartialEq)]
+pub struct Signature {
+    /// Key id
+    pub kid: Vec<u8>,
+
+    /// Algorithm used for signing
+    pub algorithm: i128,
+
+    /// Raw signature
+    pub signature: Vec<u8>,
+}
+
+/// Represents the whole certificate blob
 #[derive(Debug, PartialEq)]
 pub struct HealthCert {
     // Member country that issued the bundle (might be missing)
@@ -225,6 +244,9 @@ pub struct HealthCert {
 
     /// List of passes contained in this bundle
     pub passes: Vec<GreenPass>,
+
+    /// Raw signature
+    pub signature: Signature,
 }
 
 /// Attests the full recovery from a given disease
@@ -280,16 +302,6 @@ impl TryFrom<BTreeMap<String, Value>> for Recovery {
     }
 }
 
-/// Identifies the recognized test types
-#[derive(Debug, PartialEq)]
-pub enum TestName {
-    /// A Nucleic Acid Amplification Test, with the name of the specific test
-    NAAT { name: String }, // nm
-
-    /// A Rapid Antigen Test, with a string identifying the device from the JRC database
-    RAT { device_id: String }, // ma
-}
-
 /// Attests that a test for a given disease has been conducted.
 #[derive(Debug, PartialEq)]
 pub struct Test {
@@ -459,14 +471,34 @@ impl TryFrom<&str> for HealthCert {
         let mut data = Vec::new();
         dec.read_to_end(&mut data)?;
 
-        let Cwt(cwt_arr) = serde_cbor::from_slice(&data)?;
+        let cwt = ciborium::de::from_reader(&data[..])?;
+
+        let Cwt(cwt_arr) = cwt;
 
         if cwt_arr.len() != 4 {
             return Err(Error::MalformedCWT);
         }
 
+        let protected_properties: RawHeader = match &cwt_arr[0] {
+            Value::Bytes(_bys) => ciborium::de::from_reader(&_bys[..])?,
+            _ => {
+                return Err(Error::InvalidFormatFor {
+                    key: "protected properties".into(),
+                })
+            }
+        };
+
+        let unprotected_properties = match &cwt_arr[1] {
+            Value::Map(map) => map,
+            _ => {
+                return Err(Error::InvalidFormatFor {
+                    key: "unprotected properties".into(),
+                })
+            }
+        };
+
         let RawCert(mut cert_map) = match &cwt_arr[2] {
-            Value::Bytes(bys) => serde_cbor::from_slice(bys)?,
+            Value::Bytes(bys) => ciborium::de::from_reader(&bys[..])?,
             _ => {
                 return Err(Error::InvalidFormatFor {
                     key: "root cert".into(),
@@ -488,10 +520,10 @@ impl TryFrom<&str> for HealthCert {
         };
 
         let expires = match cert_map
-            .remove(&4)
+            .remove(&4isize)
             .ok_or_else(|| Error::MissingKey("expiration timestamp".into()))?
         {
-            Value::Integer(ts) => Utc.timestamp(ts as i64, 0),
+            Value::Integer(ts) => Utc.timestamp(i128::from(ts) as i64, 0),
             _ => {
                 return Err(Error::InvalidFormatFor {
                     key: "expiration timestamp".into(),
@@ -500,10 +532,10 @@ impl TryFrom<&str> for HealthCert {
         };
 
         let created = match cert_map
-            .remove(&6)
+            .remove(&6isize)
             .ok_or_else(|| Error::MissingKey("issue timestamp".into()))?
         {
-            Value::Integer(ts) => Utc.timestamp(ts as i64, 0),
+            Value::Integer(ts) => Utc.timestamp(i128::from(ts) as i64, 0),
             _ => {
                 return Err(Error::InvalidFormatFor {
                     key: "issue timestamp".into(),
@@ -512,7 +544,7 @@ impl TryFrom<&str> for HealthCert {
         };
 
         let hcerts = match cert_map
-            .remove(&-260)
+            .remove(&-260isize)
             .ok_or_else(|| Error::MissingKey("hcert".into()))?
         {
             Value::Map(hcmap) => hcmap
@@ -531,11 +563,66 @@ impl TryFrom<&str> for HealthCert {
             .map(GreenPass::try_from)
             .collect::<Result<Vec<_>>>()?;
 
+        let signature = match &cwt_arr[3] {
+            Value::Bytes(bys) => bys.clone(),
+            _ => {
+                return Err(Error::InvalidFormatFor {
+                    key: "signature".into(),
+                })
+            }
+        };
+
+        let mut protected_properties = protected_properties.0;
+        // The KID can be stored in the unprotected or the protected properties
+        // See https://ec.europa.eu/health/system/files/2021-04/digital-green-certificates_v3_en_0.pdf on page 7
+        // Try to get the KID from the unprotected properties
+        let kid = unprotected_properties
+            .iter()
+            .find(|&(key, _)| key == &Value::Integer(ciborium::value::Integer::from(4isize)))
+            .ok_or_else(|| Error::MissingKey("KID".into()))
+            .and_then(|kid| {
+                if let (_, Value::Bytes(bys)) = kid {
+                    Ok(bys.clone())
+                } else {
+                    Err(Error::InvalidFormatFor { key: "KID".into() })
+                }
+            });
+            
+        // If the unprotected properties don't contain a KID, try with the protected properties
+        let kid = kid.or_else(|_| {
+            match protected_properties
+                .remove(&4isize)
+                .ok_or(Error::MissingKey("KID".into()))
+            {
+                Ok(Value::Bytes(bys)) => Ok(bys),
+                _ => Err(Error::InvalidFormatFor { key: "KID".into() }),
+            }
+        })?;
+
+        let algorithm: i128 = match protected_properties
+            .remove(&1isize)
+            .ok_or_else(|| Error::MissingKey("algorithm".into()))?
+        {
+            Value::Integer(i) => i.into(),
+            _ => {
+                return Err(Error::InvalidFormatFor {
+                    key: "algorithm".into(),
+                })
+            }
+        };
+
+        let signature = Signature {
+            kid,
+            algorithm,
+            signature,
+        };
+
         Ok(HealthCert {
             some_issuer,
             created,
             expires,
             passes,
+            signature,
         })
     }
 }
